Skip to contentSkip to conversation

io_uring: add napi busy polling support

PATCH Superseded Stefan Roesch 2 participants 7 messages Apr 28, 2023 io_uring

v11 of 15 Superseded Latest is v15 →

lore
Share & tools
lore
#1 lore raw

PATCH v11 1/5 io-uring: move io_wait_queue definition to header file

In reply to Stefan Roesch

This moves the definition of the io_wait_queue structure to the header file so it can be also used from other files.

---
 io_uring/io_uring.c | 21 ---------------------
 io_uring/io_uring.h | 22 ++++++++++++++++++++++
 2 files changed, 22 insertions(+), 21 deletions(-)
io_uring/io_uring.c +0 -21
index 3d43df8f1e4e..efbd6c9c56e5 100644
@@ -2516,33 +2516,12 @@ int io_submit_sqes(struct io_ring_ctx *ctx, unsigned int nr)
 	return ret;
 }
 
-struct io_wait_queue {
-	struct wait_queue_entry wq;
-	struct io_ring_ctx *ctx;
-	unsigned cq_tail;
-	unsigned nr_timeouts;
-	ktime_t timeout;
-};
-
 static inline bool io_has_work(struct io_ring_ctx *ctx)
 {
 	return test_bit(IO_CHECK_CQ_OVERFLOW_BIT, &ctx->check_cq) ||
 	       !llist_empty(&ctx->work_llist);
 }
 
-static inline bool io_should_wake(struct io_wait_queue *iowq)
-{
-	struct io_ring_ctx *ctx = iowq->ctx;
-	int dist = READ_ONCE(ctx->rings->cq.tail) - (int) iowq->cq_tail;
-
-	/*
-	 * Wake up if we have enough events, or if a timeout occurred since we
-	 * started waiting. For timeouts, we always want to return to userspace,
-	 * regardless of event count.
-	 */
-	return dist >= 0 || atomic_read(&ctx->cq_timeouts) != iowq->nr_timeouts;
-}
-
 static int io_wake_function(struct wait_queue_entry *curr, unsigned int mode,
 			    int wake_flags, void *key)
 {
io_uring/io_uring.h +22 -0
index 25515d69d205..5f04bd47562a 100644
@@ -41,6 +41,28 @@ enum {
 	IOU_STOP_MULTISHOT	= -ECANCELED,
 };
 
+struct io_wait_queue {
+	struct wait_queue_entry wq;
+	struct io_ring_ctx *ctx;
+	unsigned cq_tail;
+	unsigned nr_timeouts;
+	ktime_t timeout;
+
+};
+
+static inline bool io_should_wake(struct io_wait_queue *iowq)
+{
+	struct io_ring_ctx *ctx = iowq->ctx;
+	int dist = READ_ONCE(ctx->rings->cq.tail) - (int) iowq->cq_tail;
+
+	/*
+	 * Wake up if we have enough events, or if a timeout occurred since we
+	 * started waiting. For timeouts, we always want to return to userspace,
+	 * regardless of event count.
+	 */
+	return dist >= 0 || atomic_read(&ctx->cq_timeouts) != iowq->nr_timeouts;
+}
+
 struct io_uring_cqe *__io_get_cqe(struct io_ring_ctx *ctx, bool overflow);
 bool io_req_cqe_overflow(struct io_kiocb *req);
 int io_run_task_work_sig(struct io_ring_ctx *ctx);
Signature · 1 line
2.39.1
#0 lore raw

v11 cover io_uring: add napi busy polling support

This adds the napi busy polling support in io_uring.c. It adds a new napi_list to the io_ring_ctx structure. This list contains the list of napi_id's that are currently enabled for busy polling. This list is used to determine which napi id's enabled busy polling. For faster access it also adds a hash table.

When a new napi id is added, the hash table is used to locate if the napi id has already been added. When processing the busy poll loop the list is used to process the individual elements.

io-uring allows specifying two parameters:
- busy poll timeout and
- prefer busy poll to call of io_napi_busy_loop()
This sets the above parameters for the ring. The settings are passed with a new structure io_uring_napi.

There is also a corresponding liburing patch series, which enables this feature. The name of the series is "liburing: add add api for napi busy poll timeout". It also contains two programs to test the this.

Testing has shown that the round-trip times are reduced to 38us from 55us by enabling napi busy polling with a busy poll timeout of 100us. More detailled results are part of the commit message of the first patch.

Changes:
- V11:
- Fixed long comment lines and whitespace issues
- Refactor new code io_cqring_wait()
- Refactor io_napi_adjust_timeout() and remove adjust_timeout
- Rename io_napi_adjust_timeout to __io_napi_adjust_timeout
- Add new function io_napi_adjust_timeout
- Cleanup calls to list_is_singular() in io_napi_multi_busy_loop() and io_napi_blocking_busy_loop()
- Cleanup io_napi_busy_loop_should_end()
- Rename __io_napi_busy_loop to __io_napi_do_busy_loop()
- V10:
- Refreshed to io-uring/for-6.4
- Repeated performance measurements for 6.4 (same/similar results)
- V9:
- refreshed to io-uring/for-6.3
- folded patch 2 and 3 into patch 4
- fixed commit description for last 2 patches
- fixed some whitespace issues
- removed io_napi_busy_loop_on helper
- removed io_napi_setup_busy helper
- renamed io_napi_end_busy_loop to io_napi_busy_loop
- removed NAPI_LIST_HEAD macro
- split io_napi_blocking_busy_loop into two functions
- added io_napi function
- comment for sqpoll check
- V8:
- added new file napi.c and add napi functions to this file
- added NAPI_LIST_HEAD function so no ifdef is necessary
- added io_napi_init and io_napi_free function
- added io_napi_setup_busy loop helper function
- added io_napi_adjust_busy_loop helper function
- added io_napi_end_busy_loop helper function
- added io_napi_sqpoll_busy_poll helper function
- some of the definitions in napi.h are macros to avoid ifdef definitions in io_uring.c, poll.c and sqpoll.c
- changed signature of io_napi_add function
- changed size of hashtable to 16. The number of entries is limited by the number of nic queues.
- Removed ternary in io_napi_blocking_busy_loop
- Rewrote io_napi_blocking_busy_loop to make it more readable
- Split off 3 more patches
- V7:
- allow unregister with NULL value for arg parameter
- return -EOPNOTSUPP if CONFIG_NET_RX_BUSY_POLL is not enabled
- V6:
- Add a hash table on top of the list for faster access during the add operation. The linked list and the hash table use the same data structure
- V5:
- Refreshed to 6.1-rc6
- Use copy_from_user instead of memdup/kfree
- Removed the moving of napi_busy_poll_to
- Return -EINVAL if any of the reserved or padded fields are not 0.
- V4:
- Pass structure for napi config, instead of individual parameters
- V3:
- Refreshed to 6.1-rc5
- Added a new io-uring api for the prefer napi busy poll api and wire it to io_napi_busy_loop().
- Removed the unregister (implemented as register)
- Added more performance results to the first commit message.
- V2:
- Add missing defines if CONFIG_NET_RX_BUSY_POLL is not defined
- Changes signature of function io_napi_add_list to static inline if CONFIG_NET_RX_BUSY_POLL is not defined
- define some functions as static

Stefan Roesch (5):
io-uring: move io_wait_queue definition to header file
io-uring: add napi busy poll support
io-uring: add sqpoll support for napi busy poll
io_uring: add register/unregister napi function
io_uring: add prefer busy poll to register and unregister napi api

 include/linux/io_uring_types.h |  11 ++
 include/uapi/linux/io_uring.h  |  12 ++
 io_uring/Makefile              |   1 +
 io_uring/io_uring.c            |  38 ++--
 io_uring/io_uring.h            |  26 +++
 io_uring/napi.c                | 315 +++++++++++++++++++++++++++++++++
 io_uring/napi.h                |  98 ++++++++++
 io_uring/poll.c                |   2 +
 io_uring/sqpoll.c              |   4 +
 9 files changed, 486 insertions(+), 21 deletions(-)
 create mode 100644 io_uring/napi.c
 create mode 100644 io_uring/napi.h

base-commit: 3c85cc43c8e7855d202da184baf00c7b8eeacf71

Signature · 1 line
2.39.1
#5 lore raw

PATCH v11 4/5 io_uring: add register/unregister napi function

This adds an api to register and unregister the napi for io-uring. If the arg value is specified when unregistering, the current napi setting for the busy poll timeout is copied into the user structure. If this is not required, NULL can be passed as the arg value.

---
 include/uapi/linux/io_uring.h | 11 ++++++++
 io_uring/io_uring.c           |  9 +++++++
 io_uring/napi.c               | 48 +++++++++++++++++++++++++++++++++++
 io_uring/napi.h               | 13 ++++++++++
 4 files changed, 81 insertions(+)
include/uapi/linux/io_uring.h +11 -0
index 0716cb17e436..278c1a9de78c 100644
@@ -523,6 +523,10 @@ enum {
 	/* register a range of fixed file slots for automatic slot allocation */
 	IORING_REGISTER_FILE_ALLOC_RANGE	= 25,
 
+	/* set/clear busy poll settings */
+	IORING_REGISTER_NAPI			= 26,
+	IORING_UNREGISTER_NAPI			= 27,
+
 	/* this goes last */
 	IORING_REGISTER_LAST,
 
@@ -649,6 +653,13 @@ struct io_uring_buf_reg {
 	__u64	resv[3];
 };
 
+/* argument for IORING_(UN)REGISTER_NAPI */
+struct io_uring_napi {
+	__u32	busy_poll_to;
+	__u32	pad;
+	__u64	resv;
+};
+
 /*
  * io_uring_restriction->opcode values
  */
io_uring/io_uring.c +9 -0
index f06175b36b41..66e4591fbe2b 100644
@@ -4405,6 +4405,15 @@ static int __io_uring_register(struct io_ring_ctx *ctx, unsigned opcode,
 			break;
 		ret = io_register_file_alloc_range(ctx, arg);
 		break;
+	case IORING_REGISTER_NAPI:
+		ret = -EINVAL;
+		if (!arg)
+			break;
+		ret = io_register_napi(ctx, arg);
+		break;
+	case IORING_UNREGISTER_NAPI:
+		ret = io_unregister_napi(ctx, arg);
+		break;
 	default:
 		ret = -EINVAL;
 		break;
io_uring/napi.c +48 -0

54 lines not on this page · raw message

io_uring/napi.h +13 -0
index 0484dfd9a396..b5e93b3777c0 100644
@@ -12,6 +12,9 @@
 void io_napi_init(struct io_ring_ctx *ctx);
 void io_napi_free(struct io_ring_ctx *ctx);
 
+int io_register_napi(struct io_ring_ctx *ctx, void __user *arg);
+int io_unregister_napi(struct io_ring_ctx *ctx, void __user *arg);
+
 void __io_napi_add(struct io_ring_ctx *ctx, struct file *file);
 
 void __io_napi_adjust_timeout(struct io_ring_ctx *ctx,
@@ -67,6 +70,16 @@ static inline void io_napi_free(struct io_ring_ctx *ctx)
 {
 }
 
+static inline int io_register_napi(struct io_ring_ctx *ctx, void __user *arg)
+{
+	return -EOPNOTSUPP;
+}
+
+static inline int io_unregister_napi(struct io_ring_ctx *ctx, void __user *arg)
+{
+	return -EOPNOTSUPP;
+}
+
 static inline bool io_napi(struct io_ring_ctx *ctx)
 {
 	return false;
Signature · 1 line
2.39.1
#4 lore raw

PATCH v11 3/5 io-uring: add sqpoll support for napi busy poll

In reply to Stefan Roesch

This adds the sqpoll support to the io-uring napi.

---
 io_uring/napi.c   | 23 +++++++++++++++++++++++
 io_uring/napi.h   |  2 ++
 io_uring/sqpoll.c |  4 ++++
 3 files changed, 29 insertions(+)
io_uring/napi.c +23 -0
index a085122cae8b..6933375e3737 100644
@@ -237,4 +237,27 @@ void __io_napi_busy_loop(struct io_ring_ctx *ctx, struct io_wait_queue *iowq)
 	}
 }
 
+/*
+ * io_napi_sqpoll_busy_poll() - busy poll loop for sqpoll
+ * @ctx: pointer to io-uring context structure
+ *
+ * Splice of the napi list and execute the napi busy poll loop.
+ */
+int io_napi_sqpoll_busy_poll(struct io_ring_ctx *ctx)
+{
+	LIST_HEAD(napi_list);
+	int ret = 0;
+
+	if (!READ_ONCE(ctx->napi_busy_poll_to))
+		return 0;
+
+	spin_lock(&ctx->napi_lock);
+	list_splice_init(&ctx->napi_list, &napi_list);
+	spin_unlock(&ctx->napi_lock);
+
+	__io_napi_do_busy_loop(&napi_list, ctx->napi_prefer_busy_poll);
+	io_napi_merge_lists(ctx, &napi_list);
+	return ret;
+}
+
 #endif
io_uring/napi.h +2 -0
index 1bdf8442081f..0484dfd9a396 100644
@@ -17,6 +17,7 @@ void __io_napi_add(struct io_ring_ctx *ctx, struct file *file);
 void __io_napi_adjust_timeout(struct io_ring_ctx *ctx,
 		struct io_wait_queue *iowq, struct timespec64 *ts);
 void __io_napi_busy_loop(struct io_ring_ctx *ctx, struct io_wait_queue *iowq);
+int io_napi_sqpoll_busy_poll(struct io_ring_ctx *ctx);
 
 static inline bool io_napi(struct io_ring_ctx *ctx)
 {
@@ -77,6 +78,7 @@ static inline void io_napi_add(struct io_kiocb *req)
 
 #define io_napi_adjust_timeout(ctx, iowq, ts) do {} while (0)
 #define io_napi_busy_loop(ctx, iowq) do {} while (0)
+#define io_napi_sqpoll_busy_poll(ctx) (0)
 
 #endif
 
io_uring/sqpoll.c +4 -0
index 9db4bc1f521a..0c8d53ef134a 100644
@@ -15,6 +15,7 @@
 #include <uapi/linux/io_uring.h>
 
 #include "io_uring.h"
+#include "napi.h"
 #include "sqpoll.h"
 
 #define IORING_SQPOLL_CAP_ENTRIES_VALUE 8
@@ -193,6 +194,9 @@ static int __io_sq_thread(struct io_ring_ctx *ctx, bool cap_entries)
 			ret = io_submit_sqes(ctx, to_submit);
 		mutex_unlock(&ctx->uring_lock);
 
+		if (io_napi(ctx))
+			ret += io_napi_sqpoll_busy_poll(ctx);
+
 		if (to_submit && wq_has_sleeper(&ctx->sqo_sq_wait))
 			wake_up(&ctx->sqo_sq_wait);
 		if (creds)
Signature · 1 line
2.39.1
#6 lore raw

PATCH v11 5/5 io_uring: add prefer busy poll to register and unregister napi api

In reply to Stefan Roesch

This adds the napi prefer busy poll setting to the register and unregister napi api. When napi is unregistered and arg is specified, both napi settings: busy poll timeout and the prefer busy poll setting are copied into the user structure.

---
 include/uapi/linux/io_uring.h |  3 ++-
 io_uring/napi.c               | 10 +++++++---
 2 files changed, 9 insertions(+), 4 deletions(-)
include/uapi/linux/io_uring.h +2 -1
index 278c1a9de78c..16d17d6ab7f7 100644
@@ -656,7 +656,8 @@ struct io_uring_buf_reg {
 /* argument for IORING_(UN)REGISTER_NAPI */
 struct io_uring_napi {
 	__u32	busy_poll_to;
-	__u32	pad;
+	__u8	prefer_busy_poll;
+	__u8	pad[3];
 	__u64	resv;
 };
 
io_uring/napi.c +7 -3
index 2d4817d1413a..5851e7eff1e8 100644
@@ -191,16 +191,18 @@ void io_napi_free(struct io_ring_ctx *ctx)
 int io_register_napi(struct io_ring_ctx *ctx, void __user *arg)
 {
 	const struct io_uring_napi curr = {
-		.busy_poll_to = ctx->napi_busy_poll_to,
+		.busy_poll_to 	  = ctx->napi_busy_poll_to,
+		.prefer_busy_poll = ctx->napi_prefer_busy_poll
 	};
 	struct io_uring_napi napi;
 
 	if (copy_from_user(&napi, arg, sizeof(napi)))
 		return -EFAULT;
-	if (napi.pad || napi.resv)
+	if (napi.pad[0] || napi.pad[1] || napi.pad[2] || napi.resv)
 		return -EINVAL;
 
 	WRITE_ONCE(ctx->napi_busy_poll_to, napi.busy_poll_to);
+	WRITE_ONCE(ctx->napi_prefer_busy_poll, !!napi.prefer_busy_poll);
 
 	if (copy_to_user(arg, &curr, sizeof(curr)))
 		return -EFAULT;
@@ -219,13 +221,15 @@ int io_register_napi(struct io_ring_ctx *ctx, void __user *arg)
 int io_unregister_napi(struct io_ring_ctx *ctx, void __user *arg)
 {
 	const struct io_uring_napi curr = {
-		.busy_poll_to = ctx->napi_busy_poll_to,
+		.busy_poll_to 	  = ctx->napi_busy_poll_to,
+		.prefer_busy_poll = ctx->napi_prefer_busy_poll
 	};
 
 	if (arg && copy_to_user(arg, &curr, sizeof(curr)))
 		return -EFAULT;
 
 	WRITE_ONCE(ctx->napi_busy_poll_to, 0);
+	WRITE_ONCE(ctx->napi_prefer_busy_poll, false);
 	return 0;
 }
 
Signature · 1 line
2.39.1
#2 lore raw

PATCH v11 2/5 io-uring: add napi busy poll support

In reply to Stefan Roesch

This adds the napi busy polling support in io_uring.c. It adds a new napi_list to the io_ring_ctx structure. This list contains the list of napi_id's that are currently enabled for busy polling. The list is synchronized by the new napi_lock spin lock. The current default napi busy polling time is stored in napi_busy_poll_to. If napi busy polling is not enabled, the value is 0.

In addition there is also a hash table. The hash table store the napi id ond the pointer to the above list nodes. The hash table is used to speed up the lookup to the list elements.

The NAPI_TIMEOUT is stored as a timeout to make sure that the time a napi entry is stored in the napi list is limited.

The busy poll timeout is also stored as part of the io_wait_queue. This is necessary as for sq polling the poll interval needs to be adjusted and the napi callback allows only to pass in one value.

This has been tested with two simple programs from the liburing library
repository: the napi client and the napi server program. The client sends a request, which has a timestamp in its payload and the server replies with the same payload. The client calculates the roundtrip time and stores it to calcualte the results.

The client is running on host1 and the server is running on host 2 (in the same rack). The measured times below are roundtrip times. They are average times over 5 runs each. Each run measures 1 million roundtrips.

no rx coal rx coal: frames=88,usecs=33
Default 57us 56us

client_poll=100us 47us 46us

server_poll=100us 51us 46us

client_poll=100us+ 40us 40us
server_poll=100us

client_poll=100us+ 41us 39us
server_poll=100us+
prefer napi busy poll on client

client_poll=100us+ 41us 39us
server_poll=100us+
prefer napi busy poll on server

client_poll=100us+ 41us 39us
server_poll=100us+
prefer napi busy poll on client + server

---
 include/linux/io_uring_types.h |  11 ++
 io_uring/Makefile              |   1 +
 io_uring/io_uring.c            |   8 ++
 io_uring/io_uring.h            |   4 +
 io_uring/napi.c                | 240 +++++++++++++++++++++++++++++++++
 io_uring/napi.h                |  83 ++++++++++++
 io_uring/poll.c                |   2 +
 7 files changed, 349 insertions(+)
 create mode 100644 io_uring/napi.c
 create mode 100644 io_uring/napi.h
include/linux/io_uring_types.h +11 -0
index 1b2a20a42413..9def1de4eaf4 100644
@@ -2,6 +2,7 @@
 #define IO_URING_TYPES_H
 
 #include <linux/blkdev.h>
+#include <linux/hashtable.h>
 #include <linux/task_work.h>
 #include <linux/bitmap.h>
 #include <linux/llist.h>
@@ -277,6 +278,16 @@ struct io_ring_ctx {
 	struct xarray		personalities;
 	u32			pers_next;
 
+#ifdef CONFIG_NET_RX_BUSY_POLL
+	struct list_head	napi_list;	/* track busy poll napi_id */
+	spinlock_t		napi_lock;	/* napi_list lock */
+
+	DECLARE_HASHTABLE(napi_ht, 4);
+	/* napi busy poll default timeout */
+	unsigned int		napi_busy_poll_to;
+	bool			napi_prefer_busy_poll;
+#endif
+
 	struct {
 		/*
 		 * We cache a range of free CQEs we can use, once exhausted it
io_uring/Makefile +1 -0
index 8cc8e5387a75..2efe7c5f07ba 100644
@@ -9,3 +9,4 @@ obj-$(CONFIG_IO_URING)		+= io_uring.o xattr.o nop.o fs.o splice.o \
 					sqpoll.o fdinfo.o tctx.o poll.o \
 					cancel.o kbuf.o rsrc.o rw.o opdef.o notif.o
 obj-$(CONFIG_IO_WQ)		+= io-wq.o
+obj-$(CONFIG_NET_RX_BUSY_POLL) += napi.o
io_uring/io_uring.c +8 -0
index efbd6c9c56e5..f06175b36b41 100644
@@ -91,6 +91,7 @@
 #include "rsrc.h"
 #include "cancel.h"
 #include "net.h"
+#include "napi.h"
 #include "notif.h"
 
 #include "timeout.h"
@@ -337,6 +338,8 @@ static __cold struct io_ring_ctx *io_ring_ctx_alloc(struct io_uring_params *p)
 	INIT_WQ_LIST(&ctx->locked_free_list);
 	INIT_DELAYED_WORK(&ctx->fallback_work, io_fallback_req_func);
 	INIT_WQ_LIST(&ctx->submit_state.compl_reqs);
+	io_napi_init(ctx);
+
 	return ctx;
 err:
 	kfree(ctx->dummy_ubuf);
@@ -2619,9 +2622,13 @@ static int io_cqring_wait(struct io_ring_ctx *ctx, int min_events,
 
 		if (get_timespec64(&ts, uts))
 			return -EFAULT;
+
 		iowq.timeout = ktime_add_ns(timespec64_to_ktime(ts), ktime_get_ns());
+		io_napi_adjust_timeout(ctx, &iowq, &ts);
 	}
 
+	io_napi_busy_loop(ctx, &iowq);
+
 	trace_io_uring_cqring_wait(ctx, min_events);
 	do {
 		unsigned long check_cq;
@@ -2856,6 +2863,7 @@ static __cold void io_ring_ctx_free(struct io_ring_ctx *ctx)
 	io_req_caches_free(ctx);
 	if (ctx->hash_map)
 		io_wq_put_hash(ctx->hash_map);
+	io_napi_free(ctx);
 	kfree(ctx->cancel_table.hbs);
 	kfree(ctx->cancel_table_locked.hbs);
 	kfree(ctx->dummy_ubuf);
io_uring/io_uring.h +4 -0
index 5f04bd47562a..d669e06f54f0 100644
@@ -48,6 +48,10 @@ struct io_wait_queue {
 	unsigned nr_timeouts;
 	ktime_t timeout;
 
+#ifdef CONFIG_NET_RX_BUSY_POLL
+	unsigned int napi_busy_poll_to;
+	bool napi_prefer_busy_poll;
+#endif
 };
 
 static inline bool io_should_wake(struct io_wait_queue *iowq)
io_uring/napi.c new file +240 -0

240 lines not on this page · raw message

io_uring/napi.h new file +83 -0

83 lines not on this page · raw message

io_uring/poll.c +2 -0
index c90e47dc1e29..0284849793bb 100644
@@ -15,6 +15,7 @@
 
 #include "io_uring.h"
 #include "refs.h"
+#include "napi.h"
 #include "opdef.h"
 #include "kbuf.h"
 #include "poll.h"
@@ -631,6 +632,7 @@ static int __io_arm_poll_handler(struct io_kiocb *req,
 		__io_poll_execute(req, mask);
 		return 0;
 	}
+	io_napi_add(req);
 
 	if (ipt->owning) {
 		/*
Signature · 1 line
2.39.1
maintainer
#3 lore raw

Re: [PATCH v11 2/5] io-uring: add napi busy poll support

This looks much better now! One question and a minor comment:

io_uring/io_uring.c static int io_cqring_wait(struct io_ring_ctx *ctx, int min_events, · line 2626

Quoted 6 lines
io_uring/io_uring.c +1 -0
@@ -2619,9 +2622,13 @@ static int io_cqring_wait(struct io_ring_ctx *ctx, int min_events,
 
 		if (get_timespec64(&ts, uts))
 			return -EFAULT;
+
 		iowq.timeout = ktime_add_ns(timespec64_to_ktime(ts), ktime_get_ns());

Probably want to kill that extra added line, not worth respinning for obviously.

Quoted 37 lines
io_uring/napi.c new file +32 -0
new file mode 100644
index 000000000000..a085122cae8b
+void __io_napi_add(struct io_ring_ctx *ctx, struct file *file)
+{
+	unsigned int napi_id;
+	struct socket *sock;
+	struct sock *sk;
+	struct io_napi_ht_entry *he;
+
+	sock = sock_from_file(file);
+	if (!sock)
+		return;
+
+	sk = sock->sk;
+	if (!sk)
+		return;
+
+	napi_id = READ_ONCE(sk->sk_napi_id);
+
+	/* Non-NAPI IDs can be rejected. */
+	if (napi_id < MIN_NAPI_ID)
+		return;
+
+	spin_lock(&ctx->napi_lock);
+	hash_for_each_possible(ctx->napi_ht, he, node, napi_id) {
+		if (he->napi_id == napi_id) {
+			he->timeout = jiffies + NAPI_TIMEOUT;
+			goto out;
+		}
+	}
+
+	he = kmalloc(sizeof(*he), GFP_NOWAIT);
+	if (!he)
+		goto out;

io_uring/napi.c line 61

Quoted 10 lines
io_uring/napi.c new file +10 -0
+
+	he->napi_id = napi_id;
+	he->timeout = jiffies + NAPI_TIMEOUT;
+	hash_add(ctx->napi_ht, &he->node, napi_id);
+
+	list_add_tail(&he->list, &ctx->napi_list);
+
+out:
+	spin_unlock(&ctx->napi_lock);
+}

Didn't look into the details here just yet, but one thing occurred to me
- would it be possible to rcu_read_lock() protect the hash for lookup? I would imagine that the ratio of successful lookups to "nope nothing found, need to alloc and insert" is quite high, and we could avoid the napi_lock for that case when just iterating the hash.

Would obviously need rcu freeing of 'he' as well, and so forth. And some way to detect if 'he' is going away or not. But seems like it'd be doable without too much trouble?

Signature · 1 line
Jens Axboe